package javax.swing.text.html.parser;  
     
import java_cup.runtime.Symbol; 
import javax.swing.text.html.HTML;
import javax.swing.text.html.HTML.Tag;       
import java.util.ArrayList;
import java.util.List;
import java.math.BigInteger; 

%%    
/* ---
		Usercode Section 
							--- */
/* ################################################################################# */


%class Lexer 	//The name of the target class
%unicode			//Full character support
%char			//Char count, used for offset reporting
%line			//Line count, used for line number reporting
%column			//Column count, used for column number reporting
%cup 			//Enables cup compatibility

%eofval{
    if (buffer.length() > 0 ) {
	    String aux = buffer.toString();
	   	buffer= new StringBuilder();
	   	return new Symbol (ParserSym.TEXT, new HTMLText(aux, offsetCDATA, false, false)); 
	}
	if(yystate()==CDATA){
		yybegin(YYINITIAL);
    	return new Symbol(ParserSym.EOF_LITERAL,currentCDATAClosingTag.substring(2,currentCDATAClosingTag.length()-1 ));
    }    
	return new Symbol (ParserSym.EOF,yytext());
%eofval}


%{	
	private StringBuilder buffer = new StringBuilder();

	private int CRcount;

	private int CRNLcount;

	private int NLcount;

	private String currentCDATAClosingTag;
	
	private int offsetCDATA=0;
	
	private LexerTextStateType textState = LexerTextStateType.DEFAULT;

	/**
	 * Decsribes whether a piece of text (no white space) was parsed.
	 */
	private boolean textParsed;
	
	/**
	 * Describes whether a line terminator has already been skipped.
	 */
	private boolean terminatorSkipped;
	
	/**
	 * Stores the position where a piece of text begins
	 */
	private int textPos;
	
	private boolean hasTrailingSpaces;
	
	private boolean hasLeadingSpaces;


	/**
	 * Indicates if the strict mode of the current Parser instance is set
	 */	 
	private boolean strict;

	/**
	 * A reference of the Parser's DTD
	 */
	private DTD dtd;
	
	/**
	 * Stores all the TAGS that are set as preformatted at <code>HTML.getAllTags()</code>
	 */
	private static List<String> preformattedTags;

	private static List<String> breaksFlowTags;

	/**
	 * This flag is used in CHECK_IF_CDATA state 
	 */
	private boolean syncronizeSent;

	/*
	 * Used by preformatted purposes. Describe if the last token was a line terminator 
	 * in order to collapse the white spaces as the RI does 
	 **/
	private boolean lastWasLineTerminator;

	/**
	 * Used by preformatted purposes. Describe if the token text returned as a SYMBOL 
	 * may have to collapse the last token (line terminator)
	 */
	private boolean skipLastLineTerminator;

/* 	private boolean lastTokenWasBreaksFlow; */

	/**
	 * A reference to the current CUP object
	 */	
	private ParserCup cup;

	/**
	 * Sets <code>preformattedTags</code> with all the tags (<code>HTML.Tag</code>) 
	 * that are set as preformatted, by examining isPreformatted() method.
	 * <br>
	 * Then, each time the method <code>isPreformatted(String)</code> is called 
	 * search at this <code>ArrayList</code>
	 */
	static {
		preformattedTags = new ArrayList<String>();
		for (Tag tag : HTML.getAllTags()) {
			if (tag.isPreformatted()) {
				preformattedTags.add(tag.toString().toLowerCase());					
			}
		}
	}

	public void setStrict(boolean strict) {
		this.strict = strict;		
	}

	public int getOffset() {
		return yychar;
	}	
	
	public void setCup (ParserCup cup){
		this.cup = cup;
	}
		
	public void setDTD(DTD dtd) {
		this.dtd = dtd;	
	}
	
	public int yyline() {
		return yyline;
	}
	
	/**
	 * Decides if a tag name corresponds to a preformatted one.
	 * <br>
     * This method use <code>preformattedTags</code> in order retrieve this information
	 * 
	 * @param tokenTag The tag name to be analyzed.
	 * @return True if the tag name corresponds to a preformatted one. Otherwise
	 *         it returns false.
	 */
	private boolean isPreformatted (String tokenTag) {
		return preformattedTags.contains(tokenTag);
	}

	private void appendText (LexerTextType textType) {
		if (buffer.length() == 0) {
			textPos = yychar;
		}
		switch (textState) {
			/*
			 * DEFAULT STATE
			 */			
			case DEFAULT:
				switch (textType)	{
					case SPACE:
					case LINE_TERMINATOR:
						if (!hasLeadingSpaces && !textParsed) {
							hasLeadingSpaces = true;
						} 
						hasTrailingSpaces = true;
					break;
					case TEXT:
						if (hasTrailingSpaces && textParsed) {
							buffer.append(" ");
						}
						buffer.append(yytext());	
						hasTrailingSpaces = false;
						textParsed = true;
					break; 
					case ENTITY_SEMI:
						if (hasTrailingSpaces && textParsed) {
							buffer.append(" ");
						}
						buffer.append(replaceEntity(true));
						textParsed = true;
						hasTrailingSpaces = false;
					break;					
					case ENTITY_NO_SEMI:
						if (hasTrailingSpaces && textParsed) {
							buffer.append(" ");
						}
						buffer.append(replaceEntity(false));
						textParsed = true;
						hasTrailingSpaces = false;
					break;
				}			
			break;
			/*
			 * PREFORMATTED STATE
			 */
			case PREFORMATTED:
				switch (textType)	{
					case SPACE:
						buffer.append(yytext());
					break;
					case LINE_TERMINATOR:
						if (terminatorSkipped) {
							buffer.append(yytext());
						}  
						/* 
						* Collapses if there is only one last line terminator. 
						* If there are more then doesn't do it (same as RI)
						* FIXME 
						*/
						skipLastLineTerminator = !lastWasLineTerminator;
						lastWasLineTerminator = true;
						terminatorSkipped = true;
					break;
					case TEXT:
						buffer.append(yytext());
						textParsed = true;
					break;
					case ENTITY_SEMI:
						buffer.append(replaceEntity(true));
						textParsed = true;
					break;					
					case ENTITY_NO_SEMI:
						buffer.append(replaceEntity(false));
						textParsed = true;
					break;
				}			
			break;
			/*
			 * CDATA STATE
			 */
			case CDATA:
				switch (textType)	{
					case TEXT:
					case SPACE:
						buffer.append(yytext()); 
						textParsed = true;
					break;
					case LINE_TERMINATOR:
						if (!terminatorSkipped) {
							// skipping line terminator 
							terminatorSkipped = true;	
						} else {
							// append line terminator
							buffer.append(yytext());
						}
					break;
					case ENTITY_SEMI:
						buffer.append(replaceEntity(true));
						textParsed = true;
					break;					
					case ENTITY_NO_SEMI:
						buffer.append(replaceEntity(false));
						textParsed = true;
					break;
				}			
			break;
		}
	}

	private String replaceEntity(boolean endsWithSemi){
        Entity entity = null;
		String str;
		if (endsWithSemi) {
        	str = yytext().substring(1, yytext().length() - 1);
        } else {
        	str = yytext().substring(1, yytext().length());
        }
                
        if (str.startsWith("#")) {
        	if (str.codePointAt(1) == 'X' || str.codePointAt(1) == 'x') {
        		entity = dtd.getEntity(Integer.parseInt(str.substring(2), 16));
        	} else {
	          	entity = dtd.getEntity(Integer.parseInt(str.substring(1)));
	        }
        } else {
           	entity = dtd.getEntity(str);
        }
        return entity == null ? "&" + str : String.valueOf(entity.data);		
	}


	private HTMLText flushBufferedText() {
		String str = buffer.toString();
		buffer = new StringBuilder();
		HTMLText textToken = null;
		if (textParsed) {
			/* 
			 * Collapses if there is only one last line terminator. 
			 * If there are more then doesn't do it (same as RI)
			 */
			if (skipLastLineTerminator && textState == LexerTextStateType.PREFORMATTED) {
				str = str.substring(0, str.length() - 1);
				skipLastLineTerminator = false;
			}
			textToken = new HTMLText(str, textPos, hasLeadingSpaces, hasTrailingSpaces);
		} 
		hasTrailingSpaces = false;
		hasLeadingSpaces = false;
		textParsed = false;
		terminatorSkipped = false;
		lastWasLineTerminator = false;
		textState = LexerTextStateType.DEFAULT;
		return textToken;
	}

	public String getEOLString() {
		int eol = Math.max(CRcount, Math.max(NLcount, CRNLcount));

		String result = null;

		if (eol == NLcount) {
			result = String.valueOf('\n');
		} else if (eol == CRNLcount) {
			result = String.valueOf("\r\n");
		} else if (eol == CRcount) {
			result = String.valueOf('\r');
		}

		if (result != null) {
			return result;
		} else {
			throw new AssertionError();
		}
	}
		
%}


/* 1 -- Character Classes */

digit			= [0-9]
lowerCaseLetter	= [a-z]
upperCaseLetter	= [A-Z]
letter			= {lowerCaseLetter}|{upperCaseLetter}
number			= [-]?{digit}+

/* 2 -- WhiteSpace Characters */

space			= " "
tab			= \t
nl			= \n
cr			= \r
crnl			= \r\n

/* 3 -- WhiteSpace */

spaceChar		= {space}|{tab}

/* 4 -- Delimiters */

COM			= "--"	/* comment */
AMP			= "&"	/* entity reference open 'ampersand' */
DQM			= \"	/* double quote mark */
SQM			= \'		/* single quote mark ' */
TAGO		= "<"	/* tag open */
TAGC		= ">"	/* tag close */
TAGOS		="</"   /* tag open with slash*/
TAGOE		= "<!"  /* tag open with exclamation*/
SLH    		= "/"	/* slash */
SMI			= ";"	/* semicolon */
SHP			= "#"	/* sharp */

//literal			= ({DQM}[^\"]*{DQM})|({SQM}[^\']*{SQM})

	

/* 5 Name Characters */

nameStartCharacter	= {lowerCaseLetter}|{upperCaseLetter}
nameCharacter		= {nameStartCharacter}|{digit}|{otherNameCharacter}
entityName			= {letter}+

otherNameCharacter	= "-"|"\."

/* 6 Names */

tagName			= {nameStartCharacter}+{nameCharacter}*
attValue 		= [^&<> \t\r\n\"\']
numberToken		= {number}{attValue}+
nameToken		= (!({number}|!{attValue}))+

hexa_init		= ("x" | "X")
hexa_tokens		= [0123456789abcdefABCDEF]+
entity_no_semi	= {AMP}({SHP}({number}|{entityName}|{hexa_init}{hexa_tokens})|{entityName})
entity_semi		= {AMP}({SHP}({number}|{entityName}|{hexa_init}{hexa_tokens})|{entityName}){SMI}

	/* A name is a name-start characer -- a letter -- followed by any number of 
	*  name characters -- letters, digits, periods, or hyphens. Entity names are case
	*  sensitive, but all other names are not.
	*
	*  FROM: http://www.w3.org/TR/WD-html-lex/
	*/

/* 7 Text Classes */

notago			= [^<]
notagc			= [^>]
noamp			= [^&]
nospace			= [^ \t\n\r]

		/* Text characters */

textChar		= !(!{notago}|!{noamp}|!{nospace}) 

text			= {textChar}+

		/* When among tags, text is anything but < or & */

mdText			= {notagc}+

		/* When in a markup declaration, the content is everything up to the first > */

commentText 	= ([^-]|-[^-]|--[^>])*

		/* When in a comment, the content is everything up to the first > (which would 
		*  be an error) or the proper commentclose: -- followed by a tagclose: >
		*/

/* ---
		Recognition Modes
							--- */
/* ################################################################################# */

/*  States */

	/* tag */
%state TAG

	/* attribute */
%state ATTR

	/* attribute value */
%state ATTRVAL

	/* end tag */
%state ENDTAG

	/* markup declaration */
%state MD

	/* comment */
%state COMMENT

	/* in-tag recognition */
%state STARTTAG	
	
	/* in-endtag recognition */
%state STARTENDTAG

	/* attribute - value split */
%state ATTREQUALS

	/* attribute value recognition */
%state ATTRVALSTART

	/* attribute literal, starting with a Double Quote Mark */
%state ATTRVALLIT_DQM

	/* attribute literal, starting with a Single Quote Mark */
%state ATTRVALLIT_SQM

	/* in-markup_declaration recognition */
%state MDSTART

	/* trailing white space*/
%state TAG_TRAILINGWS

%state CDATA

%state CDATA_ENDTAG_START

%state CDATA_ENDTAG_NAME

%state CDATA_ENDTAG_END

%state CHECK_IF_CDATA

%state YYINITIAL_NOTEXT

%state TAG_IGNORE_ATTS

%state IGNORED_Q_TAG

%%


/* ---
		Lexical Rules
							--- */
/* ################################################################################# */


/* ********************************* TAG_TRAILINGWS **************************************** */

<STARTTAG,ATTR,ATTREQUALS,ATTRVALSTART>{SLH} {
	yybegin(TAG_TRAILINGWS);
}

<TAG_TRAILINGWS>{


 {spaceChar}+ {
				yybegin(TAG_TRAILINGWS);
				}

{nl} {
	yybegin(TAG_TRAILINGWS);
	NLcount++;
}

 {cr} {
	yybegin(TAG_TRAILINGWS);
	CRcount++;
}

 {crnl} {
	yybegin(TAG_TRAILINGWS);
	CRNLcount++;
}

{TAGC} {
	yybegin(YYINITIAL);
	return new Symbol (ParserSym.TAG_SLASH_CLOSE, yytext());
}

. {
	yybegin(YYINITIAL);
	yypushback(yytext().length());	
	return new Symbol (ParserSym.TAG_SLASH_CLOSE, yytext());
}

}
/* ******************************** CDATA ************************************** */

 <CDATA> {
       
		{spaceChar}+ {
               yybegin(CDATA);
			   appendText(LexerTextType.SPACE);	
		}

  	    [^ <&\r\n]+ {
               yybegin(CDATA);
			   appendText(LexerTextType.TEXT);	
        }

        {cr} {
               yybegin(CDATA);
               CRcount++;
			   appendText(LexerTextType.LINE_TERMINATOR);				   
        }

        {nl} {
               yybegin(CDATA);
               NLcount++;
			   appendText(LexerTextType.LINE_TERMINATOR);				   
        }

        {crnl} {
			   yybegin(CDATA);
               CRNLcount++;
			   appendText(LexerTextType.LINE_TERMINATOR);				   
        }

        {entity_no_semi}  {
               yybegin(CDATA);
			   appendText(LexerTextType.ENTITY_NO_SEMI);				   
			   // to collapse next LINE TERMINATOR (same as RI)
			   terminatorSkipped = false;
        }

        {entity_semi}  {
               yybegin(CDATA);
			   appendText(LexerTextType.ENTITY_SEMI);				   
        }

       {TAGOS}{tagName}{TAGC} {
				if (yytext().equalsIgnoreCase(currentCDATAClosingTag)) {
                       	yypushback(yytext().length()-2);                       	
                       	yybegin(CDATA_ENDTAG_START);
                       	String aux = buffer.toString();
                       	
                       	if(aux.endsWith("\r\n")){
                       		aux=aux.substring(0,aux.length()-2);
                       	}
                       	else if(aux.endsWith("\n") || aux.endsWith("\r")){
                       		aux=aux.substring(0,aux.length()-1);
                       	}
                       	flushBufferedText(); // for initialization purposes
                       	return new Symbol (ParserSym.TEXT, new HTMLText(aux, offsetCDATA, false, false));                       
               } else { 
                       yybegin(CDATA);
					   appendText(LexerTextType.TEXT);				   
               }
       }            
       . {
       		yybegin(CDATA);
			appendText(LexerTextType.TEXT);				   
       }
      
      /*. {
		yybegin(YYINITIAL);
		return new Symbol (ParserSym.error,yytext());
	}*/
       
}


/* ******************************** CDATA_ENDTAG_START ************************************** */

<CDATA_ENDTAG_START> {
	{tagName} {
		yybegin(CDATA_ENDTAG_START);	
		return new Symbol (ParserSym.END_TAG_NAME, offsetCDATA, yyline, yytext()); 
	}

	{TAGC} {
		yybegin(YYINITIAL);
		return new Symbol (ParserSym.TAG_CLOSE, yychar, yyline, new Character(yytext().charAt(0)));
	}
	. {
		yybegin(CDATA_ENDTAG_START);
		return new Symbol (ParserSym.LEXERR,yytext());
	}
}


/* ******************************** YYINITIAL ************************************** */

<YYINITIAL>{
	{spaceChar}+ {
		yybegin(YYINITIAL);
		appendText(LexerTextType.SPACE);	
	} 
	
	{cr} {
		yybegin(YYINITIAL);
		appendText(LexerTextType.LINE_TERMINATOR);	
		CRcount++;
	}
	
	 {nl} {
		yybegin(YYINITIAL);
		appendText(LexerTextType.LINE_TERMINATOR);	
		NLcount++;
	}
	
	 {crnl} {
		yybegin(YYINITIAL);
		appendText(LexerTextType.LINE_TERMINATOR);	
		CRNLcount++;
	}
	
	[&] {
		yybegin(YYINITIAL);
		appendText(LexerTextType.TEXT);	
	}
	
	{text}  {
		yybegin(YYINITIAL);
		appendText(LexerTextType.TEXT);
	}
	
	{entity_no_semi}  {
		yybegin(YYINITIAL);
		appendText(LexerTextType.ENTITY_NO_SEMI);
	}
	
	{entity_semi}  { 
		yybegin(YYINITIAL);
		appendText(LexerTextType.ENTITY_SEMI);
	}

	{TAGO} {
		HTMLText textToken = flushBufferedText();
		yypushback(1);		
		yybegin(YYINITIAL_NOTEXT);
		if (textToken != null) {
			return new Symbol (ParserSym.TEXT, textToken);
		}
	}	
}
 
/* **************************** YYINITIAL_NOTEXT ************************************ */

<YYINITIAL_NOTEXT> {
	 /* <> -- empty tag*/
	{TAGO}{TAGC} {
		yybegin(YYINITIAL);
		return new Symbol (ParserSym.EMPTY_TAG, yytext());
	}

	/* </body -- end tag start*/
	{TAGO}{SLH}{tagName} {
		yybegin(ENDTAG);
		String tagName = yytext().substring(2);
		return new Symbol (ParserSym.END_TAG_NAME, yychar, yyline, tagName);
	}
	
	/* < -- tag start */
	{TAGO}{tagName} {		
		String tagName = yytext().substring(1).toLowerCase();
		Element e = dtd.elementHash.get(tagName);
		if (e != null && ((e.getType() == DTDConstants.CDATA) || e.isScript())) {
			currentCDATAClosingTag = "</" + tagName + ">";
			yybegin(TAG_IGNORE_ATTS);
		} else {
			if (isPreformatted(tagName)) {
				textState = LexerTextStateType.PREFORMATTED;
			}
			currentCDATAClosingTag = null;
			yybegin(STARTTAG);			
		}		
		return new Symbol (ParserSym.TAG_NAME, yychar, yyline, yytext().substring(1, yytext().length()));
	}	
	
	{TAGO}[?] {
		yybegin(IGNORED_Q_TAG);
	}
	
	/* <! -- tag start with exclamation*/
	{TAGOE} {
		yybegin(MD);
		return new Symbol (ParserSym.TAG_OPEN_EXM, yychar, yyline, yytext());
	}
	
	. {
		yybegin(YYINITIAL);
		return new Symbol (ParserSym.LEXERR_EXPECTED_TAG_NAME,yytext());	
	}
}

/* ***************************** IGNORED_Q_DATA ************************************ */
<IGNORED_Q_TAG>{
	[^?]+ {
	}
	[?]{TAGC} {
		yybegin(YYINITIAL);				
	}
}

/* ****************************** TAG_IGNORE_ATTS ********************************** */
/* This state was created with the purpose of ignoring tag's attributes when the actual tag is of type CDATA (RI-like behaviour) */

<TAG_IGNORE_ATTS>{
	{TAGC} {
		yybegin(CHECK_IF_CDATA);		
		syncronizeSent=false;		
		return new Symbol (ParserSym.TAG_CLOSE, yychar, yyline, new Character(yytext().charAt(0)));	
	}
	{TAGO} {
		yybegin(CHECK_IF_CDATA);
		return new Symbol (ParserSym.TAG_OPEN, new Character(yytext().charAt(0)));
	}
	{cr} {
		yybegin(TAG_IGNORE_ATTS);
		CRcount++;
	}
	
	{nl} {
		yybegin(TAG_IGNORE_ATTS);
		NLcount++;
	}
	
	{crnl} {
		yybegin(TAG_IGNORE_ATTS);
		CRNLcount++;
	}
	
	. {
	yybegin(TAG_IGNORE_ATTS);	
	}
}

/* ********************************* STARTTAG ************************************** */

<STARTTAG>{
	{TAGC} {
		yybegin(CHECK_IF_CDATA);		
		syncronizeSent=false;		
		return new Symbol (ParserSym.TAG_CLOSE, yychar, yyline, new Character(yytext().charAt(0)));	
	}
	{TAGO} {
		yybegin(CHECK_IF_CDATA);
		return new Symbol (ParserSym.TAG_OPEN, new Character(yytext().charAt(0)));
	}
	{spaceChar}+ {
		yybegin(ATTR);
	} 
	
	{cr} {
		yybegin(ATTR);
		CRcount++;
	}
	
	{nl} {
		yybegin(ATTR);
		NLcount++;
	}
	
	{crnl} {
		yybegin(ATTR);
		CRNLcount++;
	}
	
	[:] {
		yybegin(TAG_IGNORE_ATTS);
		return new Symbol (ParserSym.LEXERR,yytext());
	}
	
	
	. {
		yybegin(STARTTAG);
		return new Symbol (ParserSym.LEXERR,yytext());
	}
}
 
 
/* ********************************** ENDTAG *************************************** */
 
<ENDTAG>{

	{spaceChar}+ {
		yybegin(ENDTAG);
	} 
	
	{cr} {
		yybegin(ENDTAG);
		CRcount++;
	}
	
	{nl} {
		yybegin(ENDTAG);
		NLcount++;
	}
	
	{crnl} {
		yybegin(ENDTAG);
		CRNLcount++;
	}
	
	{TAGC} {
		yybegin(YYINITIAL);
		return new Symbol (ParserSym.TAG_CLOSE, yychar, yyline, new Character(yytext().charAt(0)));
	}	
	. {
		yybegin(ENDTAG);
		return new Symbol (ParserSym.LEXERR,yytext());
	}	
}

/* *********************************** ATTR **************************************** */

<ATTR>{
	{tagName} {
		yybegin(ATTREQUALS);
		return new Symbol (ParserSym.ATTR_NAME, yytext());
	}
	
	 {spaceChar}+ {
		yybegin(ATTR);
	} 
	
	 {cr} {
		yybegin(ATTR);
		CRcount++;
	}
	
	 {nl} {
		yybegin(ATTR);
		NLcount++;
	}
	
	 {crnl} {
		yybegin(ATTR);
		CRNLcount++;
	}
	
	{TAGC} {
		yybegin(YYINITIAL);
		return new Symbol (ParserSym.TAG_CLOSE, yychar, yyline, new Character(yytext().charAt(0)));
	}
	
	. {
		yybegin(ATTR);		
		return new Symbol (ParserSym.LEXERR,yytext());
	}
}


/* ********************************** ATTREQUALS *********************************** */

<ATTREQUALS>{
	
 	{tagName} {
		yybegin(ATTREQUALS);
		return new Symbol (ParserSym.ATTR_NAME, yytext());
	}

	{spaceChar}+ {
		yybegin(ATTREQUALS);
	} 
	
	{cr} {
		yybegin(ATTREQUALS);
		CRcount++;
	}
	
	{nl} {
		yybegin(ATTREQUALS);
		NLcount++;
	}
	
	{crnl} {
		yybegin(ATTREQUALS);
		CRNLcount++;
	}
	
	"=" {
		yybegin(ATTRVAL);
		return new Symbol (ParserSym.EQUALS, new Character(yytext().charAt(0)));
	}
	
	{TAGC} {
		yybegin(YYINITIAL);
		return new Symbol (ParserSym.TAG_CLOSE, yychar, yyline, new Character(yytext().charAt(0)));
	}	
	. {
		yybegin(ATTREQUALS);
		return new Symbol (ParserSym.LEXERR,yytext());
	}
	
}

/* ********************************** ATTRVAL ************************************** */

<ATTRVAL>{
	{spaceChar}+ {
		yybegin(ATTRVAL);
	} 
	
	 {cr} {
		yybegin(ATTRVAL);
		CRcount++;
	}
	
	 {nl} {
		yybegin(ATTRVAL);
		NLcount++;
	}
	
	 {crnl} {
		yybegin(ATTRVAL);
		CRNLcount++;
	}
	
	{TAGC} {
		yybegin(YYINITIAL);
		return new Symbol (ParserSym.TAG_CLOSE, yychar, yyline, new Character(yytext().charAt(0)));
	}
	
	/* Single quote marks can be included within the attribute value when the value is 
	*  delimited by double quote marks, and viceversa.
	*
	*  FROM: W3C's HTML 3.2 Reference Specification
	*/
	
	
	{DQM} {
		yybegin(ATTRVALLIT_DQM);
		return new Symbol (ParserSym.DQM, new Character(yytext().charAt(0)));
	}
	
	{SQM} {
		yybegin(ATTRVALLIT_SQM);
		return new Symbol (ParserSym.SQM, new Character(yytext().charAt(0)));
	}
	
}	

<ATTRVAL, ATTRVALSTART>{

	{entity_no_semi}  {
		yybegin(ATTRVALSTART);
		return new Symbol (ParserSym.ATTRVAL_NMTK, replaceEntity(false));
	}
	
	{entity_semi}  {
		yybegin(ATTRVALSTART);
		return new Symbol (ParserSym.ATTRVAL_NMTK, replaceEntity(true));
	}
	
	/* The case where a single '&' symbol is found */
	{AMP} {
		yybegin(ATTRVALSTART);
		return new Symbol (ParserSym.ATTRVAL_NMTK, yytext());
	}
	

	{number} {
		yybegin(ATTRVALSTART);
		return new Symbol (ParserSym.ATTRVAL_NUM, yytext());
	}
	
	{numberToken} {
		yybegin(ATTRVALSTART);
		return new Symbol (ParserSym.ATTRVAL_NUMTK, yytext());
	}
	
	{nameToken} {
		yybegin(ATTRVALSTART);
		return new Symbol (ParserSym.ATTRVAL_NMTK, yytext());
	}
	/*. {
		yybegin(ATTRVALSTART);
		return new Symbol (ParserSym.LEXERR,yytext());
	}*/
}

/* ********************************* ATTRVALLIT_DQM ************************************ */

<ATTRVALLIT_DQM>{
	{spaceChar}+ {
		yybegin(ATTRVALLIT_DQM);
	}
	
	 {cr} {
		yybegin(ATTRVALLIT_DQM);
		CRcount++;
	}
	
	 {nl} {
		yybegin(ATTRVALLIT_DQM);
		NLcount++;
	}
	
	 {crnl} {
		yybegin(ATTRVALLIT_DQM);
		CRNLcount++;
	}
	
	{DQM} {
		yybegin(ATTR);
		return new Symbol (ParserSym.DQM, new Character(yytext().charAt(0)));
	}
	
	{entity_semi}  {
		yybegin(ATTRVALLIT_DQM);
		return new Symbol (ParserSym.ATTRVAL_LIT, replaceEntity(true));
	}
	
	{entity_no_semi}  {
		yybegin(ATTRVALLIT_DQM);
		return new Symbol (ParserSym.ATTRVAL_LIT, replaceEntity(false));
	}
	
	/* The case where a single '&' symbol is found */
	{AMP} {
		yybegin(ATTRVALLIT_DQM);
		return new Symbol (ParserSym.ATTRVAL_LIT, yytext());
	}
	
	[^\"&]*  {
		yybegin(ATTRVALLIT_DQM);
		return new Symbol (ParserSym.ATTRVAL_LIT, yytext());
	}
}

/* ********************************* ATTRVALLIT_SQM ************************************ */

<ATTRVALLIT_SQM>{
	{spaceChar}+ {
		yybegin(ATTRVALLIT_SQM);
	}
	
	 {cr} {
		yybegin(ATTRVALLIT_SQM);
		CRcount++;
	}
	
	 {nl} {
		yybegin(ATTRVALLIT_SQM);
		NLcount++;
	}
	
	 {crnl} {
		yybegin(ATTRVALLIT_SQM);
		CRNLcount++;
	}
	
	{SQM} {
		yybegin(ATTR);
		return new Symbol (ParserSym.SQM, new Character(yytext().charAt(0)));
	}
	
	{entity_semi}  {
		yybegin(ATTRVALLIT_SQM);
		return new Symbol (ParserSym.ATTRVAL_LIT, replaceEntity(true));
	}
	
	{entity_no_semi}  {
		yybegin(ATTRVALLIT_SQM);
		return new Symbol (ParserSym.ATTRVAL_LIT, replaceEntity(false));
	}
	
	/* The case where a single '&' symbol is found */
	{AMP} {
		yybegin(ATTRVALLIT_SQM);
		return new Symbol (ParserSym.ATTRVAL_LIT, yytext());
	}
	
	[^\'&]*  {
		yybegin(ATTRVALLIT_SQM);
		return new Symbol (ParserSym.ATTRVAL_LIT, yytext());
	}
}


/* ******************************* ATTRVALSTART *********************************** */

<ATTRVALSTART>{
	{TAGC} {
		yybegin(YYINITIAL);
		return new Symbol (ParserSym.TAG_CLOSE, yychar, yyline, new Character(yytext().charAt(0)));
	}
	
	{DQM} { 
		yybegin(ATTRVALSTART);
		return new Symbol (ParserSym.DQM, new Character(yytext().charAt(0)));
	}
	
	 {spaceChar}+ {
		yybegin(ATTR);
	} 
	
	 {cr} {
		yybegin(ATTR);
		CRcount++;
	}
	
	 {nl} {
		yybegin(ATTR);
		NLcount++;
	}
	
	 {crnl} {
		yybegin(ATTR);
		CRNLcount++;
	}
	. {
		yybegin(ATTRVALSTART);
		return new Symbol (ParserSym.LEXERR,yytext());
	}
	
}



/* ********************************** MD ****************************************** */

/* <!DOCTYPE -- markup declaration */

<MD>{
	{tagName} {
		yybegin(MDSTART);
		return new Symbol (ParserSym.MUDECL, yytext());
	}
	
	/* <!--  -- comment declaration */
	{COM} {
		yybegin(COMMENT);
		return new Symbol (ParserSym.COMM, yytext());
	}
	
	
	 {spaceChar}+ {
		yybegin(MD);
	} 
	
	 {cr} {
		yybegin(MD);
		CRcount++;
	}
	
	 {nl} {
		yybegin(MD);
		NLcount++;
	}
	
	 {crnl} {
		yybegin(MD);
		CRNLcount++;
	}
	. {
		yybegin(MD);
		return new Symbol (ParserSym.LEXERR,yytext());
	}
}

/* ********************************** MDSTART ************************************** */

<MDSTART>{
	{TAGC} {
		yybegin(YYINITIAL);
		return new Symbol (ParserSym.TAG_CLOSE, yychar, yyline, new Character(yytext().charAt(0)));
	}	
	{mdText} {
		yybegin(MDSTART);
		return new Symbol (ParserSym.MUDECL_CONTENT, yytext());
	}

}

/* *********************************** COMMENT ************************************* */

<COMMENT>{
	{commentText} {
		yybegin(COMMENT);
		return new Symbol (ParserSym.COMM_CONTENT, yytext());
	}	
	{COM}{TAGC} {
		yybegin(YYINITIAL);
		return new Symbol (ParserSym.TAG_COMM_CLOSE, yytext());
	}
	. {
		yybegin(COMMENT);
		return new Symbol (ParserSym.LEXERR,yytext());
	}
}


/* ************************************ CHECK IF CDATA ***************************** */

<CHECK_IF_CDATA>{
	.|\n {		
		yypushback(yytext().length());
		
		if(!syncronizeSent){
			syncronizeSent=true;
			return new Symbol (ParserSym.SYNC, yytext());			
		}
				
		if (cup.action_obj.isLastStartTagCreatedOk() && currentCDATAClosingTag!=null) {
			textState = LexerTextStateType.CDATA;
			yybegin(CDATA);
		} else {			
			yybegin(YYINITIAL);
		}
	}
}